import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from .dimensionalty_reduction import DimensionalityReduction


class Pca(DimensionalityReduction):

    def __init__(self):
        super().__init__()
        self.name = 'pca'

    @staticmethod
    def requires_all_samples():
        return True

    def compute(self, similarity_ts, ts2_filename):
        assert len(similarity_ts.ts1_windows) > 1, \
            'TS1 was not split. PCA needs more than 1 TS1 windows. Check TS1 size and stride parameter.'
        super().compute(similarity_ts, ts2_filename)
        pca = PCA(n_components=2)
        pca.fit(self.ts1_reduced_dimensions)
        pca_ts1 = pca.transform(self.ts1_reduced_dimensions)
        pca_ts2 = pca.transform(self.ts2_reduced_dimensions)
        fig, axis = self.__generate_plot(pca_ts1, pca_ts2)
        return [(fig, axis)]

    def __generate_plot(self, pca_ts1, pca_ts2):
        fig, axis = super()._init_plot()
        n_samples_ts1 = self.ts1_windows.shape[0]
        colors = super()._generate_colors(len(pca_ts1), len(pca_ts2))
        plt.scatter(pca_ts1[:, 0], pca_ts1[:, 1],
                    c=colors[:n_samples_ts1], alpha=0.2, label='TS_1')
        plt.scatter(pca_ts2[:, 0], pca_ts2[:, 1],
                    c=colors[n_samples_ts1:], alpha=0.2, label='TS_2')
        super()._set_labels('PCA', 'x_pca', 'y_pca')
        plt.close('all')
        return fig, axis



import numpy as np
from .metric import Metric


class Cc(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'cc'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__cc(ts1, ts2)}
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__cc(ts1, ts2)

    def __cc(self, ts1, ts2):
        ts1_covariance = np.cov(ts1)
        generated_data_covariance = np.cov(ts2)
        covariance_diff_matrix = ts1_covariance - generated_data_covariance
        l1_norms_avg = np.mean([np.linalg.norm(row) for row in covariance_diff_matrix])
        return l1_norms_avg



import itertools
import matplotlib.pyplot as plt
import numpy as np
from sklearn.manifold import TSNE
from .dimensionalty_reduction import DimensionalityReduction


class Tsne(DimensionalityReduction):

    def __init__(self):
        super().__init__()
        self.name = 'tsne'

    @staticmethod
    def requires_all_samples():
        return True

    def compute(self, similarity_ts, ts2_filename):
        super().compute(similarity_ts, ts2_filename)
        ts1_ts2_reduced_dimensions_concatenated = np.concatenate(
            (self.ts1_reduced_dimensions, self.ts2_reduced_dimensions), axis=0)
        perplexities = [5, 10, min(40, len(self.ts1_reduced_dimensions))]
        perplexities = [perplexity for perplexity in perplexities if perplexity <= len(self.ts1_reduced_dimensions)]
        iterations = [300, 1000]
        plot_array = []
        for perplexity, n_iterations in itertools.product(perplexities, iterations):
            tsne_embedding = self.__compute_tsne(ts1_ts2_reduced_dimensions_concatenated, n_iterations, perplexity)
            tsne_plot = self.__generate_plot(tsne_embedding, n_iterations, perplexity)
            plot_array.append(tsne_plot)
        return plot_array

    def __compute_tsne(self, ts1_ts2_reduced_dimensions_concatenated, n_iterations, perplexity):
        tsne = TSNE(n_components=2, verbose=0, perplexity=perplexity, n_iter=n_iterations)
        return tsne.fit_transform(ts1_ts2_reduced_dimensions_concatenated)

    def __generate_plot(self, tsne_embedding, n_iterations, perplexity):
        fig, axis = super()._init_plot()
        n_samples_ts1 = self.ts1_reduced_dimensions.shape[0]
        colors = super()._generate_colors(len(self.ts1_reduced_dimensions), len(self.ts2_reduced_dimensions))
        plt.scatter(tsne_embedding[:n_samples_ts1, 0], tsne_embedding[:n_samples_ts1, 1],
                    c=colors[:n_samples_ts1], alpha=0.2, label='TS_1')
        plt.scatter(tsne_embedding[n_samples_ts1:, 0], tsne_embedding[n_samples_ts1:, 1],
                    c=colors[n_samples_ts1:], alpha=0.2, label='TS_2')
        super()._set_labels(f't-SNE-iter_{n_iterations}-perplexity_{perplexity}', 'x_tsne', 'y_tsne')
        plt.close('all')
        return fig, axis



from matplotlib import pyplot as plt


class Plot:
    @staticmethod
    def requires_all_samples():
        return False

    def __init__(self, fig_size=(18, 3)):
        self.ts2_filename = None
        self.ts1 = None
        self.ts2 = None
        self.ts1_windows = None
        self.header_names = None
        self.fig_size = fig_size

    def _initialize(self, similarity_ts, ts2_filename):
        self.ts2_filename = ts2_filename
        self.ts1 = similarity_ts.ts1_ts2_associated_windows[self.ts2_filename]['most_similar_ts1_sample']
        self.ts2 = similarity_ts.ts1_ts2_associated_windows[self.ts2_filename]['ts2']
        self.ts1_windows = similarity_ts.ts1_windows
        self.header_names = similarity_ts.header_names

    def compute(self, similarity_ts, ts2_filename):
        self._initialize(similarity_ts, ts2_filename)

    def _init_plot(self):
        plt.rcParams['figure.figsize'] = self.fig_size
        return plt.subplots(1)

    def _set_labels(self, title, x_label, y_label):
        plt.title(title)
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.legend(loc='best')



import numpy as np
from scipy.spatial import cKDTree as KDTree
from .metric import Metric


class Kl(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'kl'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.kl_divergence(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update({column_name: self.kl_divergence_univariate(ts1[:, column_index].reshape(-1, 1),
                                                                                    ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.kl_divergence(ts1, ts2)

    @staticmethod
    def kl_divergence_univariate(array_1, array_2, range_values=None, num_bins=10):
        # smoothing implemented as per https://www.cs.bgu.ac.il/~elhadad/nlp09/KL.html
        # pc = eps*|SU-SP|/|SP| and qc = eps*|SU-SQ|/|SQ|.
        # eps=0.0001
        # SP and SQ the samples observed in P and Q respectively
        # SU = SP U SQ
        # pc = eps*|SU-SP|/|SP| and qc = eps*|SU-SQ|/|SQ|.
        # P'(i) = P(i) - pc if i in SP
        # P'(i) = eps otherwise for i in SU - SP
        eps = 0.000001
        min_all = min(array_1.min(), array_2.min())
        max_all = max(array_1.max(), array_2.max())
        range_values = range_values if range_values is not None else (min_all, max_all)
        p = np.histogram(array_1, bins=np.linspace(range_values[0], range_values[1], num_bins + 1))[0] / len(array_1)
        q = np.histogram(array_2, bins=np.linspace(range_values[0], range_values[1], num_bins + 1))[0] / len(array_2)
        pc = eps * (num_bins - (p != 0).sum()) / (p != 0).sum()
        pq = eps * (num_bins - (q != 0).sum()) / (q != 0).sum()
        p = np.vectorize(lambda p_i: eps if p_i == 0 else p_i - pc)(p)
        q = np.vectorize(lambda q_i: eps if q_i == 0 else q_i - pq)(q)
        kl_p_m = sum(p[i] * np.log(p[i] / q[i]) for i in range(len(p)))
        kl_q_m = sum(q[i] * np.log(q[i] / p[i]) for i in range(len(p)))
        return kl_p_m, kl_q_m

    @staticmethod
    def kl_divergence(x, y):
        """Compute the Kullback-Leibler divergence between two multivariate samples.

        Parameters
        ----------
        x : 2D array (n,d)
            Samples from distribution P, which typically represents the true
            distribution.
        y : 2D array (m,d)
            Samples from distribution Q, which typically represents the approximate
            distribution.

        Returns
        -------
        out : float
            The estimated Kullback-Leibler divergence D(P||Q).

        References
        ----------
        Pérez-Cruz, F. Kullback-Leibler divergence estimation of
        continuous distributions IEEE International Symposium on Information
        Theory, 2008.
        """
        x = np.atleast_2d(x)
        y = np.atleast_2d(y)
        n, d = x.shape
        m, dy = y.shape
        assert d == dy, 'Both samples must be consistent in shape.'
        # Build a KD tree representation of the samples and find the nearest neighbour
        # of each point in x.
        xtree = KDTree(x)
        ytree = KDTree(y)
        eps = 0.000001
        # Get the first two nearest neighbours for x, since the closest one is the
        # sample itself.
        r = xtree.query(x, k=2, eps=.01, p=2)[0][:, 1]
        s = ytree.query(x, k=1, eps=.01, p=2)[0]
        pr = eps * (len(r) - (r != 0).sum()) / (r != 0).sum()
        ps = eps * (len(s) - (s != 0).sum()) / (s != 0).sum()
        r = np.vectorize(lambda r_i: eps if r_i == 0 else r_i - pr)(r)
        s = np.vectorize(lambda s_i: eps if s_i == 0 else s_i - ps)(s)
        # There is a mistake in the paper. In Eq. 14, the right side misses a negative sign
        # on the first term of the right hand side.
        return -np.log(r / s).sum() * d / n + np.log(m / (n - 1.))



import numpy as np
from .metric import Metric


class Hi(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'hi'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__compute_hi(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update(
                {column_name: self.__compute_hi(ts1[:, column_index].reshape(-1, 1), ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__compute_hi(ts1, ts2)

    def __compute_hi(self, ts1, ts2):
        histogram_diff_matrix = []
        for column in range(0, ts1.shape[1]):
            ts1_column_values = ts1[:, column]
            ts1_histogram, _ = np.histogram(ts1_column_values)
            generated_data_column_values = ts2[:, column]
            generated_histogram, _ = np.histogram(generated_data_column_values)
            column_histogram_diff = ts1_histogram - generated_histogram
            histogram_diff_matrix.append(column_histogram_diff)
        histogram_diff_matrix = np.asmatrix(histogram_diff_matrix)
        l1_norms_histogram_diff = np.apply_along_axis(np.linalg.norm, 1, histogram_diff_matrix)

        l1_norms_histogram_diff_avg = l1_norms_histogram_diff.mean()

        return l1_norms_histogram_diff_avg



from sklearn import metrics
from .metric import Metric


class Mmd(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'mmd'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__mmd_calculate_rbf(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update({column_name: self.__mmd_calculate_rbf(ts1[:, column_index].reshape(-1, 1),
                                                                               ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__mmd_calculate_rbf(ts1, ts2)

    def __mmd_calculate_rbf(self, x, y, gamma=1.0):
        xx = metrics.pairwise.rbf_kernel(x, x, gamma)
        yy = metrics.pairwise.rbf_kernel(y, y, gamma)
        xy = metrics.pairwise.rbf_kernel(x, y, gamma)
        return xx.mean() + yy.mean() - 2 * xy.mean()



from .metric_factory import MetricFactory


class MetricConfig:
    def __init__(self, metrics_names_to_be_computed=None):
        if metrics_names_to_be_computed is None:
            metrics_names_to_be_computed = MetricFactory.find_available_metrics().keys()
        self.metrics = metrics_names_to_be_computed



import numpy as np
from .plot import Plot


class DimensionalityReduction(Plot):
    def __init__(self, fig_size=(8, 6)):
        super().__init__(fig_size)
        self.ts1_reduced_dimensions = None
        self.ts2_reduced_dimensions = None

    def _initialize(self, similarity_ts, ts2_filename):
        super()._initialize(similarity_ts, ts2_filename)
        self.ts1_reduced_dimensions, self.ts2_reduced_dimensions = self.__reduce_tss_dimensionality(
            similarity_ts.ts1_windows,
            np.asarray(similarity_ts.ts2s))

    def __reduce_tss_dimensionality(self, ts1, ts2):
        ts1_reduced = self.__reduce_ts_dimensionality(ts1)
        ts2_reduced = self.__reduce_ts_dimensionality(ts2)
        return ts1_reduced, ts2_reduced

    def __reduce_ts_dimensionality(self, ts):
        n_samples, n_steps, n_features = ts.shape
        return np.reshape(ts, (n_samples, n_steps * n_features))

    def _generate_colors(self, color1_size, color2_size):
        return ['red' for _ in range(color1_size)] + ['blue' for _ in range(color2_size)]



from dtaidistance import dtw_visualisation
from dtaidistance import dtw
from matplotlib import pyplot as plt
from .plot import Plot


class Dtw(Plot):

    def __init__(self):
        super().__init__()
        self.name = 'dtw'

    def compute(self, similarity_ts, ts2_filename):
        super().compute(similarity_ts, ts2_filename)
        figures = []
        for column_name, index in zip(self.header_names, range(len(self.header_names))):
            figures.append(
                self.__generate_plot(
                    self.ts1[:, index], self.ts2[:, index], column_name
                )
            )
        return figures

    def __generate_plot(self, ts1_column, ts2_column, column_name):
        super()._init_plot()
        path = dtw.warping_path(ts1_column, ts2_column)
        fig, axes = dtw_visualisation.plot_warping(ts1_column, ts2_column, path)
        plt.xlim(left=0, right=len(ts1_column) - 1)
        axes[0].set_title(f'DTW_{column_name}')
        axes[0].legend(['TS_1'], loc='center right')
        axes[1].legend(['TS_2'], loc='center right')
        plt.close('all')
        return fig, axes



import matplotlib.pyplot as plt
import pandas
from .plot import Plot


class TwoDimensions(Plot):
    def __init__(self):
        super().__init__()
        self.seq_len = None
        self.ts2_df = None
        self.ts1_df = None
        self.name = '2d'

    def _initialize(self, similarity_ts, ts2_filename):
        super()._initialize(similarity_ts, ts2_filename)
        self.ts1_df = pandas.DataFrame(self.ts1, columns=[f'{column_name}_TS_1' for column_name in
                                                          similarity_ts.header_names])
        self.ts2_df = pandas.DataFrame(similarity_ts.ts2_dict[ts2_filename],
                                       columns=[f'{column_name}_TS_2' for column_name in
                                                similarity_ts.header_names])
        self.seq_len = similarity_ts.ts2_dict[ts2_filename].shape[0]

    def compute(self, similarity_ts, ts2_filename):
        super().compute(similarity_ts, ts2_filename)
        plot_array = [self.__generate_plot_from_df()]
        for column_index, column_name in enumerate(self.header_names):
            plot_array.append(
                self.__generate_plot_by_column(self.ts1[:, column_index], self.ts2[:, column_index], column_name))
        return plot_array

    def __generate_plot_from_df(self):
        fig, axis = super()._init_plot()
        self.ts1_df.plot(ax=axis, style='--')
        plt.gca().set_prop_cycle(None)
        self.ts2_df.plot(ax=axis)
        plt.xlim(left=0, right=len(self.ts2_df) - 1)
        super()._set_labels('complete_TS_1_vs_TS_2', 'time', 'values')
        plt.close('all')
        return fig, axis

    def __generate_plot_by_column(self, ts1_column, ts2_column, column_name):
        fig, axis = super()._init_plot()
        plt.plot(ts1_column, c='green', label='TS_1', linewidth=1)
        plt.plot(ts2_column, c='blue', label='TS_2', linewidth=2)
        plt.xlim(left=0, right=len(ts1_column) - 1)
        super()._set_labels(f'{column_name}_TS_1_vs_TS_2', 'time', column_name)
        plt.close('all')
        return fig, axis



import numpy as np
from .metric import Metric


class Cp(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'cp'

    def compute(self, ts1, ts2, similarity_ts):
        return {'Multivariate': self.__cp(ts1, ts2)}

    def compute_distance(self, ts1, ts2):
        return self.__cp(ts1, ts2)

    def __cp(self, ts1, ts2):
        ts1_pearson = np.corrcoef(ts1, rowvar=False)
        ts2_pearson = np.corrcoef(ts2, rowvar=False)
        correlation_diff_matrix = ts1_pearson - ts2_pearson
        l1_norms_avg = np.mean([np.linalg.norm(row) for row in correlation_diff_matrix])
        return l1_norms_avg



from .metrics.metric_config import MetricConfig
from .plots.plot_config import PlotConfig


class SimilarityTsConfig:
    def __init__(self, metric_config=None, plot_config=None, stride=1, window_selection_metric='dtw', ts2_names=None,
                 header_names=None):
        if metric_config is None:
            metric_config = MetricConfig()
        if plot_config is None:
            plot_config = PlotConfig()

        self.metric_config = metric_config
        self.plot_config = plot_config
        self.stride = stride
        self.window_selection_metric = window_selection_metric
        self.ts2_names = ts2_names
        self.header_names = header_names



import os
from .metric import Metric
from ..helpers.dynamic_import_helper import find_available_classes


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class MetricFactory(metaclass=Singleton):
    def __init__(self, metrics_names_to_be_computed):
        self.metric_objects = self.__get_metrics_to_be_computed(metrics_names_to_be_computed)

    def register_metric(self, metric_class):
        new_metric = metric_class()
        self.metric_objects.append(new_metric)

    @staticmethod
    def __get_metrics_to_be_computed(metrics_names_to_be_computed):
        return [metric for metric_name, metric in MetricFactory.find_available_metrics().items() if
                metric_name in metrics_names_to_be_computed]

    @staticmethod
    def find_available_metrics():
        return find_available_classes(os.path.dirname(os.path.abspath(__file__)), Metric, 'metrics')

    @staticmethod
    def get_metric_by_name(metric_name):
        return MetricFactory.find_available_metrics()[metric_name]

    @staticmethod
    def get_instance(metrics_to_be_computed=None):
        if not hasattr(MetricFactory, '_instance'):
            if metrics_to_be_computed is None:
                metrics_to_be_computed = MetricFactory.find_available_metrics().keys()
            MetricFactory._instance = MetricFactory(metrics_to_be_computed)
        return MetricFactory._instance



import os
from .plot import Plot
from ..helpers.dynamic_import_helper import find_available_classes


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class PlotFactory(metaclass=Singleton):
    def __init__(self, figure_names_to_be_generated):
        self.plots_to_be_generated = self.__get_plots_to_be_generated(figure_names_to_be_generated)
        self.figures_requires_all_samples = self.__get_figures_that_requires_all_samples()

    def register_plot(self, plot_class):
        new_plot = plot_class()
        self.plots_to_be_generated.append(new_plot)

    @staticmethod
    def __get_plots_to_be_generated(figure_names_to_be_generated):
        return [plot for plot_name, plot in PlotFactory.find_available_figures().items() if
                plot_name in figure_names_to_be_generated]

    @staticmethod
    def find_available_figures():
        return find_available_classes(os.path.dirname(os.path.abspath(__file__)), Plot, 'plots')

    @staticmethod
    def get_instance(figure_names_to_be_generated=None):
        if not hasattr(PlotFactory, '_instance'):
            if figure_names_to_be_generated is None:
                figure_names_to_be_generated = PlotFactory.find_available_figures().keys()
            PlotFactory._instance = PlotFactory(figure_names_to_be_generated)
        return PlotFactory._instance

    def __get_figures_that_requires_all_samples(self):
        return [plot.name for plot in self.plots_to_be_generated if
                plot.requires_all_samples()]



import importlib
import os


def find_available_classes(folder_path, parent_class, package):
    available_classes = {}
    for _, _, files in os.walk(folder_path):
        for file_name in files:
            if file_name.endswith('.py'):
                module_name = file_name[:-3]
                module = importlib.import_module(f'similarity_ts.{package}.{module_name}', package=package)
                if __is_module_subclass(module, module_name, parent_class):
                    available_class = getattr(module, __to_camel_case(module_name))()
                    available_classes[available_class.name] = available_class
    return available_classes


def __is_module_subclass(module, module_name, parent_class):
    if module is not None:
        class_name = __to_camel_case(module_name)
        if parent_class.__name__ != module_name.capitalize():
            return hasattr(module, class_name) and issubclass(getattr(module, class_name), parent_class)
    return False



def __to_camel_case(snake_str):
    components = snake_str.split('_')
    return ''.join(x.capitalize() for x in components)



import csv
import os
import chardet
import numpy as np
from natsort import natsorted


def __detect_encoding(file_path):
    with open(file_path, 'rb') as file:
        raw_data = file.read()
        result = chardet.detect(raw_data)
        return result['encoding'], result['confidence']


def __read_header_from_csv(filename, ts_delimiter, has_header):
    if has_header:
        header = np.genfromtxt(filename, delimiter=ts_delimiter, names=has_header, max_rows=1, dtype=str).dtype.names
    else:
        first_row = np.loadtxt(filename, delimiter=ts_delimiter, max_rows=1)
        header = ['column-' + str(i) for i in range(len(first_row))]
    return header


def __detect_line_delimiter(filename):
    encoding, _ = __detect_encoding(filename)
    with open(filename, 'r', newline='', encoding=encoding) as file:
        ts_delimiter = csv.Sniffer().sniff(file.readline()).delimiter

    return ts_delimiter


def load_ts_from_csv(filename, has_header=None):
    ts_delimiter = __detect_line_delimiter(filename)

    header = __read_header_from_csv(filename, ts_delimiter, has_header)
    skiprows = 1 if has_header else 0

    return np.loadtxt(filename, delimiter=ts_delimiter, skiprows=skiprows), header


def load_ts_from_path(path, header_ts1, has_header=None):
    if not os.path.exists(path):
        raise FileNotFoundError(f' Path {path} does not exist')
    time_series = {}
    if os.path.isfile(path):
        ts2, header_ts2 = load_ts_from_csv(path, has_header)
        __check_headers(header_ts1, header_ts2)
        time_series[os.path.basename(path)] = ts2
    elif os.path.isdir(path):
        for _, _, files in os.walk(path):
            files = [file for file in files if not file.startswith('.')]
            files = natsorted(files)
            for file in files:
                ts2, header_ts2 = load_ts_from_csv(f'{path}/{file}', has_header)
                __check_headers(header_ts1, header_ts2)
                time_series[file] = ts2

    return time_series


def __check_headers(header_ts1, header_ts2):
    if header_ts1 != header_ts2:
        raise ValueError('All time series must have the same header column names.')



from .similarity_ts_config import SimilarityTsConfig
from .metrics.metric_computer import MetricComputer
from .metrics.metric_factory import MetricFactory
from .plots.plot_computer import PlotComputer
from .plots.plot_factory import PlotFactory
from .helpers.window_sampler import create_ts1_ts2_associated_windows, split_ts_strided


class SimilarityTs:
    def __init__(self, ts1, ts2s, similarity_ts_config=None):
        self.ts1 = ts1
        self.ts2s = ts2s
        self.similarity_ts_config = similarity_ts_config if similarity_ts_config is not None else SimilarityTsConfig()
        self.header_names = self.similarity_ts_config.header_names if self.similarity_ts_config.header_names is not None else [
            'column-' + str(i)
            for i in
            range(ts1.shape[1])]
        self.ts2_dict = self.__build_ts2_dict(self.ts2s, self.similarity_ts_config.ts2_names)
        self.ts1_windows = split_ts_strided(self.ts1, self.ts2s[0].shape[0], self.similarity_ts_config.stride)
        self.ts1_ts2_associated_windows = create_ts1_ts2_associated_windows(self)
        self.metric_factory = MetricFactory.get_instance(self.similarity_ts_config.metric_config.metrics)
        self.plot_factory = PlotFactory.get_instance(self.similarity_ts_config.plot_config.figures)


    def __build_ts2_dict(self, ts2s, ts2_filenames):
        ts2_filenames = ts2_filenames if ts2_filenames is not None else ['ts2_' + str(i) for i in range(len(ts2s))]
        return {ts2_name: ts2 for ts2, ts2_name in zip(ts2s, ts2_filenames)}

    def get_metric_computer(self):
        return MetricComputer(self, self.metric_factory.metric_objects)

    def get_plot_computer(self):
        return PlotComputer(self, self.plot_factory.plots_to_be_generated)



import warnings
from ..similarity_analysis_computer import SimilarityAnalysisComputer


class PlotComputer(SimilarityAnalysisComputer):
    def __init__(self, similarity_ts, analysis):
        super().__init__(similarity_ts, analysis)
        self.already_computed_figures_requires_all_samples = []

    def _compute_next_analysis(self):
        plot = next(self.analysis_iterator)
        ts2_filename, _ = self.current_associated_window
        computed_plots = []
        if plot.name not in self.already_computed_figures_requires_all_samples:
            try:
                if plot.requires_all_samples():
                    self.already_computed_figures_requires_all_samples.append(plot.name)
                computed_plots = plot.compute(self.similarity_ts, ts2_filename)
            except Exception as e:
                warnings.warn(f'Warning: Plot {plot.name} could not be computed. Details: {e}', Warning)
        return ts2_filename, plot.name, computed_plots



import warnings
from ..similarity_analysis_computer import SimilarityAnalysisComputer


class MetricComputer(SimilarityAnalysisComputer):

    def _compute_next_analysis(self):
        metric = next(self.analysis_iterator)
        filename, ts_dict = self.current_associated_window
        if metric.name in ts_dict['cached_metric'].keys():
            computed_metric = ts_dict['cached_metric'][metric.name]
        else:
            try:
                computed_metric = metric.compute(ts_dict['most_similar_ts1_sample'], ts_dict['ts2'], self.similarity_ts)
            except Exception as e:
                computed_metric = []
                warnings.warn(f'\nWarning: Metric {metric.name} could not be computed. Details: {e}', Warning)
        return filename, metric.name, computed_metric



class Metric:

    def compute(self, ts1, ts2, similarity_ts):
        raise NotImplementedError('Subclasses must implement compute() method')

    def compute_distance(self, ts1, ts2):
        raise NotImplementedError('Subclasses must implement compute_distance() method')

    def compare(self, metric1, metric2):
        return metric2 - metric1



import statistics
import scipy
from .metric import Metric

class Ks(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'ks'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__compute_ks(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update({column_name: self.__compute_ks(ts1[:, column_index].reshape(-1, 1), ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__compute_ks(ts1, ts2)

    def __compute_ks(self, ts1, ts2):
        column_indexes = range(ts2.shape[1])
        return statistics.mean(
            [scipy.stats.ks_2samp(ts2[:, column_index], ts1[:, column_index])[0] for
            column_index in column_indexes])



from itertools import cycle
import matplotlib.pyplot as plt
import numpy as np
from .plot import Plot


class Delta(Plot):

    def __init__(self):
        super().__init__()
        self.seq_len = None
        self.ts_freq_secs = None
        self.n_ts1_samples_to_plot = None
        self.time_magnitude = None
        self.time_magnitude_name = None
        self.name = 'delta'

    def _initialize(self, similarity_ts, ts2_filename):
        super()._initialize(similarity_ts, ts2_filename)
        self.seq_len = similarity_ts.ts2_dict[ts2_filename].shape[0]
        self.ts_freq_secs = similarity_ts.similarity_ts_config.plot_config.timestamp_frequency_seconds
        self.n_ts1_samples_to_plot = min(5, len(self.ts1_windows))
        self.time_magnitude, self.time_magnitude_name = self.__compute_time_magnitude()

    def __compute_time_magnitude(self):
        time_magnitudes = {60 * 60 * 24: 'days', 60 * 60: 'hours', 60: 'minutes'}
        selected_magnitude, selected_name = 1, 'seconds'
        for magnitude, name in time_magnitudes.items():
            if self.ts_freq_secs >= magnitude:
                selected_magnitude, selected_name = magnitude, name
                break
        return selected_magnitude, selected_name

    def compute(self, similarity_ts, ts2_filename):
        super().compute(similarity_ts, ts2_filename)
        time_intervals = [(self.ts_freq_secs / self.time_magnitude) * value for value in [2, 5, 10]]
        plot_array = []
        for column_index, column_name in enumerate(self.header_names):
            for time_interval in time_intervals:
                plot_array.append(
                    self.__generate_delta_grouped_by_interval(time_interval, column_index, column_name))
        return plot_array

    def __generate_delta_grouped_by_interval(self, time_interval, column_index, column_name):
        delta_ts1_column_array = [
            self.__compute_grouped_delta_from_sample(self.__get_random_ts1_sample(), column_index, time_interval) for _
            in
            range(self.n_ts1_samples_to_plot)]

        delta_ts2_column = self.__compute_grouped_delta_from_sample(self.ts2, column_index, time_interval)

        return self.__generate_plot(delta_ts1_column_array=delta_ts1_column_array,
                                    delta_ts2_column=delta_ts2_column, column_name=column_name,
                                    time_interval=time_interval)

    def __get_random_ts1_sample(self):
        return self.ts1_windows[np.random.randint(0, len(self.ts1_windows))]

    def __compute_grouped_delta_from_sample(self, data_sample, column_number, time_interval):
        sample_column = data_sample[:, column_number]
        seq_len = data_sample.shape[0]
        sample_column_split = np.array_split(sample_column,
                                             seq_len // (time_interval / (self.ts_freq_secs / self.time_magnitude)))
        sample_column_mean = [np.mean(batch) for batch in sample_column_split]
        delta_sample_column = -np.diff(sample_column_mean)
        return delta_sample_column

    def __generate_plot(self, delta_ts1_column_array, delta_ts2_column, column_name, time_interval):
        fig, axis = super()._init_plot()
        self.__plot_ts1_columns(delta_ts1_column_array)
        plt.plot(delta_ts2_column, c='blue', label='TS_2', linewidth=3)
        max_y = max(np.amax(delta_ts1_column_array), np.amax(delta_ts2_column))
        min_y = min(np.amin(delta_ts1_column_array), np.amin(delta_ts2_column))
        axis_limits = [0, len(delta_ts2_column) - 1, min_y, max_y]
        plt.axis(axis_limits)
        super()._set_labels(f'{column_name}_TS_1_vs_TS_2_(grouped_by_{int(time_interval)}_{self.time_magnitude_name})',
                            'time', column_name)
        plt.close('all')
        return fig, axis

    def __plot_ts1_columns(self, delta_ts1_column_array):
        cycle_colors = cycle('grcmk')
        labels = ['TS_1'] if len(delta_ts1_column_array) == 1 else [f'TS_1_sample_{i}' for i in
                                                                    range(1, len(delta_ts1_column_array) + 1)]
        for label, delta_ts1_column in zip(labels, delta_ts1_column_array):
            plt.plot(delta_ts1_column, c=next(cycle_colors), label=label, linewidth=1)



from .plot_factory import PlotFactory


class PlotConfig:
    def __init__(self, figure_names_to_be_generated=None, timestamp_frequency_seconds=1):
        assert timestamp_frequency_seconds > 0, 'Timestamp frequency seconds must be greater than 0'
        if figure_names_to_be_generated is None:
            figure_names_to_be_generated = PlotFactory.find_available_figures().keys()
        self.figures = figure_names_to_be_generated
        self.timestamp_frequency_seconds = timestamp_frequency_seconds



import numpy as np
from dtaidistance import dtw_ndim
from .metric import Metric


class Dtw(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'dtw'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__compute_dtw(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update(
                {column_name: self.__compute_dtw(ts1[:, column_index].reshape(-1, 1), ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__compute_dtw(ts1, ts2)

    def __compute_dtw(self, ts1, ts2):
        sample_length = len(ts2)
        processed_ts2 = np.insert(ts2, 0, np.ones(sample_length, dtype=int), axis=1)
        processed_ts2 = np.insert(processed_ts2, 0, range(sample_length), axis=1)
        processed_ts1 = np.insert(ts1, 0, np.ones(sample_length, dtype=int), axis=1)
        processed_ts1 = np.insert(processed_ts1, 0, range(sample_length), axis=1)

        return dtw_ndim.distance_fast(processed_ts2, processed_ts1)



class SimilarityAnalysisComputer:
    def __init__(self, similarity_ts, analysis):
        self.ts1_ts2_associated_windows_iterator = iter(similarity_ts.ts1_ts2_associated_windows.items())
        self.similarity_ts = similarity_ts
        self.analysis = analysis
        self.analysis_iterator = iter(analysis)
        self.length = len(analysis) * len(similarity_ts.ts1_ts2_associated_windows.items())
        self.current_associated_window = None

    def __iter__(self):
        self.current_associated_window = next(self.ts1_ts2_associated_windows_iterator)
        return self

    def __next__(self):
        try:
            return self._compute_next_analysis()
        except StopIteration:
            self.current_associated_window = next(self.ts1_ts2_associated_windows_iterator)
            self.analysis_iterator = iter(self.analysis)
            return self._compute_next_analysis()

    def __len__(self):
        return self.length

    def _compute_next_analysis(self):
        raise NotImplementedError('Subclasses must implement _compute_next_analysis() method')



import numpy as np
from tqdm import tqdm
from ..metrics.metric_factory import MetricFactory


def __get_most_similar_ts_sample(ts1_windows, ts2, metric_object):
    current_best = float('inf')
    most_similar_sample = []
    for ts1_window in ts1_windows:
        current_distance = metric_object.compute_distance(ts1_window, ts2)
        if metric_object.compare(current_distance, current_best) > 0:
            current_best = current_distance
            most_similar_sample = ts1_window
    return most_similar_sample


def split_ts_strided(ts_np, seq_len, stride):
    assert seq_len <= ts_np.shape[0], 'Seq_len cannot be greater than the original dataset length.'
    assert (ts_np.shape[
                0] - seq_len) >= stride - 1, 'Stride cannot be greater than the size difference between time series.'
    start_sequence_range = list(range(0, (ts_np.shape[0] - seq_len) + 1, stride))
    ts_windows = np.array(
        [ts_np[start_index:start_index + seq_len] for start_index in start_sequence_range])
    return ts_windows


def create_ts1_ts2_associated_windows(similarity_ts):
    metric_object = MetricFactory.get_metric_by_name(similarity_ts.similarity_ts_config.window_selection_metric)
    ts1_ts2_associated_windows = {}
    for filename, ts2 in tqdm(similarity_ts.ts2_dict.items(), desc='Selecting most similar windows'):
        most_similar_ts1_sample = __get_most_similar_ts_sample(similarity_ts.ts1_windows, ts2, metric_object)
        cached_metric = {similarity_ts.similarity_ts_config.window_selection_metric: metric_object.compute(most_similar_ts1_sample, ts2, similarity_ts)}
        ts1_ts2_associated_windows[filename] = {}
        ts1_ts2_associated_windows[filename]['most_similar_ts1_sample'] = most_similar_ts1_sample
        ts1_ts2_associated_windows[filename]['ts2'] = ts2
        ts1_ts2_associated_windows[filename]['cached_metric'] = cached_metric
    return ts1_ts2_associated_windows



from .kl import Kl
from .metric import Metric


class Js(Metric):

    def __init__(self):
        super().__init__()
        self.name = 'js'

    def compute(self, ts1, ts2, similarity_ts):
        metric_result = {'Multivariate': self.__js_distance_multivariate(ts1, ts2)}
        for column_name, column_index in zip(similarity_ts.header_names, range(len(similarity_ts.header_names))):
            metric_result.update(
                {column_name: self.__js_distance(ts1[:, column_index].reshape(-1, 1), ts2[:, column_index].reshape(-1, 1))})
        return metric_result

    def compute_distance(self, ts1, ts2):
        return self.__js_distance_multivariate(ts1, ts2)

    def __js_distance(self, ts1, ts2, num_bins=100):
        kl_p_m, kl_q_m = Kl.kl_divergence_univariate(ts1, ts2, num_bins=num_bins)
        return (kl_p_m + kl_q_m) / 2

    def __js_distance_multivariate(self, ts1, ts2):
        kl_divergence_1 = Kl.kl_divergence(ts1, ts2)
        kl_divergence_2 = Kl.kl_divergence(ts2, ts1)
        return (kl_divergence_1 + kl_divergence_2) / 2
